#!/usr/bin/env python3
"""
Port Forwarder
TCP/UDP port forwarding utility.
Parameters:
    - local_port: Local port to listen on
    - remote_host: Remote host to forward to
    - remote_port: Remote port to forward to
    - protocol: Protocol to use (tcp/udp)
Requires: admin/root privileges for privileged ports
"""
import socket
import select
import threading
import argparse
import logging
from datetime import datetime

class PortForwarder:
    def __init__(self, local_port, remote_host, remote_port, protocol='tcp'):
        self.local_port = local_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.protocol = protocol.lower()
        self.connections = []
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def handle_tcp_client(self, client_socket):
        try:
            remote_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            remote_socket.connect((self.remote_host, self.remote_port))
            
            while True:
                readable, _, _ = select.select([client_socket, remote_socket], [], [])
                
                for sock in readable:
                    other = remote_socket if sock is client_socket else client_socket
                    data = sock.recv(4096)
                    
                    if not data:
                        return
                    other.send(data)
        except Exception as e:
            self.logger.error(f"Error handling TCP connection: {str(e)}")
        finally:
            client_socket.close()
            remote_socket.close()

    def handle_udp_client(self, data, addr, server):
        try:
            remote_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            remote_socket.sendto(data, (self.remote_host, self.remote_port))
            
            response, _ = remote_socket.recvfrom(4096)
            server.sendto(response, addr)
        except Exception as e:
            self.logger.error(f"Error handling UDP packet: {str(e)}")
        finally:
            remote_socket.close()

    def start_tcp_forward(self):
        try:
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server.bind(('0.0.0.0', self.local_port))
            server.listen(5)
            
            self.logger.info(f"TCP Port forwarder listening on port {self.local_port}")
            self.logger.info(f"Forwarding to {self.remote_host}:{self.remote_port}")
            
            while True:
                client, addr = server.accept()
                self.logger.info(f"Accepted TCP connection from {addr[0]}:{addr[1]}")
                
                client_thread = threading.Thread(
                    target=self.handle_tcp_client,
                    args=(client,)
                )
                client_thread.daemon = True
                client_thread.start()
                self.connections.append(client_thread)
                
        except Exception as e:
            self.logger.error(f"Error in TCP forwarder: {str(e)}")
        finally:
            server.close()

    def start_udp_forward(self):
        try:
            server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            server.bind(('0.0.0.0', self.local_port))
            
            self.logger.info(f"UDP Port forwarder listening on port {self.local_port}")
            self.logger.info(f"Forwarding to {self.remote_host}:{self.remote_port}")
            
            while True:
                data, addr = server.recvfrom(4096)
                self.logger.info(f"Received UDP packet from {addr[0]}:{addr[1]}")
                
                client_thread = threading.Thread(
                    target=self.handle_udp_client,
                    args=(data, addr, server)
                )
                client_thread.daemon = True
                client_thread.start()
                self.connections.append(client_thread)
                
        except Exception as e:
            self.logger.error(f"Error in UDP forwarder: {str(e)}")
        finally:
            server.close()

    def start_forward(self):
        if self.protocol == 'tcp':
            self.start_tcp_forward()
        elif self.protocol == 'udp':
            self.start_udp_forward()
        else:
            self.logger.error(f"Unsupported protocol: {self.protocol}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Port Forwarder")
    parser.add_argument("local_port", type=int, help="Local port to listen on")
    parser.add_argument("remote_host", help="Remote host to forward to")
    parser.add_argument("remote_port", type=int, help="Remote port to forward to")
    parser.add_argument("--protocol", choices=['tcp', 'udp'], default='tcp',
                      help="Protocol to use (default: tcp)")
    args = parser.parse_args()

    forwarder = PortForwarder(args.local_port, args.remote_host, 
                             args.remote_port, args.protocol)
    forwarder.start_forward()
